#####################

"""
Name: SlopeClassScript.py
 Description: A script that iteratively simplifies the slope classes of a slope raster and converts
              the simplified raster to an attributed polygon feature class.
 Author: Ryan Stanley, WVU GIS Technical Center
 Email: rjstanley@mix.wvu.edu
 Date: January 2013

 ArcGIS Tool Notes:
       The script parameters must be set properly in the ArcGIS script tool
       for the script to run.  This script should be bundled with an
       ArcGIS toolbox (.tbx) that contains a tool linked to this script. To run this
       script via an ArcGIS toolbox, the tool parameters must be set as shown below 

       Python variable         Tool Display Name                       Data Type
       ---------------         -----------------                       ---------
       slopeRasterPath         Select Slope Raster                     Raster Dataset
       reclassValues           Set Slope Raster Reclass Values         Remap
       minDelineationSize      Input Minimum Delineation Size (acres)  Double
       geoDatabasePath         Select Geodatabase for Output           Workspace or Feature Dataset
       deleteTempFiles         Delete Temporary Files?                 Boolean
       NA (output)             Output Feature Class                    Feature Class


 ArcGIS Tool Validator Notes:
       The Remap datatype has a bug in ArcGIS 10 so custom tools from scripts
       do not register and display the default remap value table. The workaround
       requires that the tool validation function 'initializeParameters' be modified so the
       Remap tool parameter is set to [0].


       def initializeParameters(self):
           self.params[1].parameterDependencies = [0]
           return
"""


import arcpy
from arcpy import env
from arcpy.sa import *
import os

class SlopeClass:
    
    def __init__(self, slopeRasterPath, reclassValues, minDelineationSize, geoDatabasePath, deleteTempFiles):

        # Set class instance variables equal to input parameters
        self.slopeRasterPath = slopeRasterPath
        self.reclassValues = self.convertReclassValues(reclassValues)
        self.minDelineationSize = minDelineationSize
        self.geoDatabasePath = geoDatabasePath
        self.deleteTempFiles = bool(deleteTempFiles)

        # Temporary file names will be appended to this list
        self.temporaryFiles = []

        # Check that the 3D Analyst and Spatial Analyst extentions are available before running script
        self.checkArcGISExtentions()

        # Make arcy raster object from slope raster path
        rasterType = arcpy.Describe(self.slopeRasterPath).dataType
        if rasterType == 'RasterLayer':
            self.slopeRaster = arcpy.Raster(str(self.slopeRasterPath.dataSource))
        elif rasterType == 'RasterDataset':
            self.slopeRaster = arcpy.Raster(str(self.slopeRasterPath))
            arcpy.CalculateStatistics_management(str(self.slopeRasterPath))
        else:
            self.slopeRaster = arcpy.Raster(self.slopeRasterPath)
            arcpy.CalculateStatistics_management(str(self.slopeRasterPath))
            

        # Get spatial reference from slope raster
        self.sr = arcpy.Describe(self.slopeRaster).spatialReference

        # set scratch workspace
        env.workspace = self.geoDatabasePath
        env.scratchWorkspace =  self.geoDatabasePath
        env.overwriteOutput = True

        # The checkRasterProjection function ensures the slope raster is projected.
        # The script stops if the raster only has a GCS or no coordinate systeam at all,
        # else it returns the linear units of the projected coordinate system
        linearUnits = self.checkRasterProjection()

        # Get slope raster cell size in the linear units of the coordinate system
        self.rasterCellSize = float(self.slopeRaster.meanCellHeight)

        # Get max cluster size (in pixels) based upon linear units (feet or meters)
        maxClusterSize = self.getMaxClusterSize(self.rasterCellSize, linearUnits)  

        # The getClusterSizeList function returns a list with three cluster sizes, used in Stage 1
        self.clusterCellSizeList = self.getClusterSizeList(maxClusterSize)

        # Create the query to select raster groups with number of cells greater than the cluster size
        self.maxClusterSizeQuery = "Count > {0}".format(maxClusterSize)


    def run(self):
        arcpy.AddMessage('Running Stage 1...')

        # Run Stage1 function
        stageOneRaster = self.Stage1()
        #stageOneRaster = "Stage1_FinalRaster"

        # Get new slope class values from output of Stage1 function
        slopeClassValuesList = self.getSlopeClassList(stageOneRaster)

        arcpy.AddMessage(' ')
        arcpy.AddMessage('Running Stage 2...')
                         
        # Run Stage2 function
        stageTwoRaster = self.Stage2(stageOneRaster, slopeClassValuesList)
        #stageTwoRaster = "MOSAIC_6"

        arcpy.AddMessage(' ')
        arcpy.AddMessage('Running Stage 3...')
        
        # Run Stage3 function
        outputFeatureClass = self.Stage3(stageTwoRaster)

        # Delete the necessary files
        mustDelete = ['BoundaryClean1', 'BoundaryClean2', 'ExtractRaster', 'ExtractRaster2', 'MajorityFilter1', 'NibbleRaster', 'RegionGroup']
        for tempFile in mustDelete:
            try:
                arcpy.Delete_management(tempFile)
            except:
                pass

        # Delete temporary files
        if self.deleteTempFiles is True:
            
            arcpy.AddMessage('Deleting Temporary Files...')
            for tempFile in self.temporaryFiles:
                try:
                    arcpy.Delete_management(tempFile)
                except:
                    pass
        else:
            pass

        # Add final slope polygon feature class to ArcGIS display
        arcpy.SetParameter(5, outputFeatureClass)

    # Checks if ArcGIS extensions are available
    def checkArcGISExtentions(self):
        try:
            if arcpy.CheckExtension("3D") == "Available":
                arcpy.CheckOutExtension("3D")
            else:
                arcpy.AddError('The 3D extension is not available')
                raise Exception
            
            if arcpy.CheckExtension("Spatial") == "Available":
                arcpy.CheckOutExtension("Spatial")
            else:
                arcpy.AddError('The Sptial Analyst extension is not available')
                raise Exception
        except:
            print arcpy.GetMessages(2)
            arcpy.AddError("Please ensure the 3d analyst and spatial analyst extentions are enabled.")
            raise Exception

    # Splits the reclass value string into a dictonary for joining to attribute table in Stage3
    def splitReclassValues(self, reclassValues):
        reclassDict = {}
        valuesString = str(reclassValues).split(';')
        valuesList = list(valuesString)
        for values in valuesList:
            vsplit = values.split()
            reclassDict[vsplit[2]] = (vsplit[0],vsplit[1])
        return reclassDict

    # Alters the user-submitted reclass list so the upper limit is equal 8000, runs on class initialization
    def convertReclassValues(self, reclassValues):
        reclassSplit = reclassValues.split(';')
        
        # modify list numbers so they are sequential starting at 1
        reclassList = [[int(s.split()[0]), int(s.split()[1]), i+1] for i,s in enumerate(reclassSplit)]

        # change the upper limit of the last reclass value to 8000
        reclassList[-1][1] = 8000

        # Convert python reclass list into arcpy RemapRange object
        remapList = RemapRange(reclassList)
        return remapList

    # Gets raster cell values (VAT) and converts into list of sequential values, runs after Stage1  
    def getSlopeClassList(self,stageOneRaster):
        rasterValues = [row.Value for row in arcpy.SearchCursor(stageOneRaster)]
        rLen = len(rasterValues)
        slopeClassValuesList = [[rasterValues[i],rasterValues[i+1]] for i,x in enumerate(rasterValues) if i != rLen-1]
        return slopeClassValuesList

    # Runs zonal statistics on final raster using slope polygons, runs in Stage3, returns dictonary that relates
    # polygon features to raster cell values based upon majority of cells
    def getZonalStats(self, polygonFeatureLayer, finalRasterLayer):
        zonalStatsTable = r'in_memory\zonalstat'
        arcpy.PolygonToRaster_conversion(polygonFeatureLayer, "UNIQUEID", "ZoneRaster", cellsize = self.rasterCellSize)
        arcpy.BuildRasterAttributeTable_management("ZoneRaster", "Overwrite")
       
        ZonalStatisticsAsTable("ZoneRaster", "Value", finalRasterLayer, zonalStatsTable, "DATA", "MAJORITY")

        zoneDict = {}
        rows = arcpy.SearchCursor(zonalStatsTable)
        for row in rows:
            zoneDict[row.Value] = row.MAJORITY

        self.temporaryFiles.append("ZoneRaster")
        return zoneDict

    # Checks projection of slope raster.  Fails if raster is not projected or only has a GCS, else returns linear unit
    def checkRasterProjection(self):
        srType = self.sr.type
       
        if srType == 'Geographic':
            arcpy.AddError("The slope raster has only a geographic coordinate system. \
                            Please project the raster into a projected coordinate system and try again.")
            raise Exception
        
        elif srType == 'Projected':
            linearUnits = self.sr.linearUnitName
            return linearUnits

        elif srType != 'Projected' and srType != 'Geographic':
            arcpy.AddError("The slope raster does not have a projection. Please project the \
                            raster into a projected coordinate system and try again.")
            raise Exception

    # Gets max raster cell cluster size given min delineation size, raster cell size, and linear units
    def getMaxClusterSize(self, rasterCellSize, linearUnits):

        # Calculate the maximum cluster size based upon the linear units (feet or meters)
        if linearUnits == 'Feet' or linearUnits == 'Foot' or linearUnits == 'Foot_US':
            maxClusterSize = self.minDelineationSize/(rasterCellSize**2/43560)
        elif linearUnits == 'Meter':
            maxClusterSize = self.minDelineationSize/(rasterCellSize**2/4047)
        else:
            arcpy.AddError("The slope raster is projected in a coordinate system that does not use feet \
                            or meters as the linear unit.  Please project the slope raster and try again.")
            raise Exception

        return maxClusterSize
            
    # Get cluster size list used for three passes on Stage 1
    def getClusterSizeList(self, maxClusterSize):
        percentSizes = [0.0125, 0.050, 0.125]
        clusterSizeList = [round(x*maxClusterSize) for x in percentSizes]
        return clusterSizeList

    # Runs Stage 1 of slope script, call 9 geoprocessing tasks
    def Stage1(self):

        # First delete any 'in memory' raster data
        arcpy.Delete_management('in_memory')

        # Reclassify slope raster using submitted reclass values
        reclassRaster = Reclassify(self.slopeRaster, "Value", self.reclassValues)
        reclassRaster.save('Reclassified_Slope')
        arcpy.DefineProjection_management('Reclassified_Slope', self.sr)
        arcpy.CalculateStatistics_management('Reclassified_Slope')

        arcpy.AddMessage(str(arcpy.ListRasters()))

        # First Boundary Clean
        boundaryClean1 = BoundaryClean('Reclassified_Slope', "DESCEND", "TWO_WAY")
        boundaryClean1.save("BoundaryClean1")
        arcpy.CalculateStatistics_management("BoundaryClean1")

        arcpy.AddMessage(str(arcpy.ListRasters()))

        # Second Boundary Clean
        boundaryClean2 = BoundaryClean("BoundaryClean1", "DESCEND", "TWO_WAY")
        boundaryClean2.save("BoundaryClean2")
        arcpy.CalculateStatistics_management("BoundaryClean2")

        arcpy.AddMessage(str(arcpy.ListRasters()))
            
        majorityFilter1 = MajorityFilter("BoundaryClean2", "FOUR", "HALF")
        majorityFilter1.save('MajorityFilter1')
        arcpy.CalculateStatistics_management('MajorityFilter1')
        
        majorityFilter2 = MajorityFilter('MajorityFilter1', "FOUR", "HALF")
        majorityFilter2.save("MajorityFilter2_Stage1")
        arcpy.CalculateStatistics_management("MajorityFilter2_Stage1")
        arcpy.BuildRasterAttributeTable_management("MajorityFilter2_Stage1", "Overwrite")

        self.temporaryFiles.append("MajorityFilter2_Stage1")
            
        # Set nibbleOutput to majority filter 2 raster, will be updated each loop hereafter
        nibbleOutput = "MajorityFilter2_Stage1"

        # Loop for Stage 1 function
        counter = 1
        for clusterCellSize in self.clusterCellSizeList:
            message = "Stage 1, Pass {0} of 3".format(counter)
            arcpy.AddMessage(message)
            counter += 1
            
            queryValue = "Count > {0}".format(clusterCellSize)
            regionGroup = RegionGroup(nibbleOutput, "FOUR", "WITHIN", "NO_LINK")
            regionGroup.save('RegionGroup')
            arcpy.AddMessage('RegionGroup success')
            arcpy.CalculateStatistics_management('RegionGroup')
            
            extractRaster = ExtractByAttributes('RegionGroup', queryValue)
            extractRaster.save('ExtractRaster')
            arcpy.CalculateStatistics_management('ExtractRaster')
            
            nibbleRaster = Nibble(nibbleOutput, 'ExtractRaster')
            nibbleRaster.save('NibbleRaster')
            arcpy.CalculateStatistics_management('NibbleRaster')
            arcpy.BuildRasterAttributeTable_management('NibbleRaster', "Overwrite")
            
            nibbleOutput = 'NibbleRaster'

        # Save rasters and define projections
        arcpy.CopyRaster_management(nibbleOutput, "Stage1_FinalRaster")
        arcpy.DefineProjection_management("Stage1_FinalRaster", self.sr)
        arcpy.CalculateStatistics_management("Stage1_FinalRaster")
        arcpy.BuildRasterAttributeTable_management("Stage1_FinalRaster", "Overwrite")

        return "Stage1_FinalRaster"

    # Runs Stage 2 of slope script
    def Stage2(self, stageOneRaster, slopeClassValuesList):
       
        numSlopeRanges = len(slopeClassValuesList)
        slopeClasses = numSlopeRanges + 1
        arcpy.AddMessage("The reclassified slope raster contains " + str(slopeClasses) + \
                         " slope classes after completing Stage 1.")

        # Set mosaic raster to final nibbled output from stage 1, will be updated each loop
        MosaicRaster = stageOneRaster
        arcpy.CalculateStatistics_management(MosaicRaster)

        # Counter variable updated +1 during each loop through slope class list
        counter = 1
        
        # Loop through slope class values list
        for adjacentSlopeClassValues in slopeClassValuesList:
            message = "Stage 2, Pass {0} of {1}".format(counter, numSlopeRanges)
            arcpy.AddMessage(message)

            queryValue = "Value = {0} or Value = {1}".format(adjacentSlopeClassValues[0], adjacentSlopeClassValues[1])
            extractRaster = ExtractByAttributes(MosaicRaster, queryValue)
            extractRaster.save('ExtractRaster')
            arcpy.CalculateStatistics_management('ExtractRaster')
            
            regionGroup = RegionGroup('ExtractRaster', "FOUR", "WITHIN", "NO_LINK")
            regionGroup.save('RegionGroup')
            arcpy.CalculateStatistics_management('RegionGroup')
            
            extractRaster2 = ExtractByAttributes('RegionGroup', self.maxClusterSizeQuery)
            extractRaster2.save('ExtractRaster2')
            arcpy.CalculateStatistics_management('ExtractRaster2')
            
            nibbleRaster = Nibble('ExtractRaster', 'ExtractRaster2')
            nibbleRaster.save('NibbleRaster')
            arcpy.CalculateStatistics_management('NibbleRaster')
            arcpy.BuildRasterAttributeTable_management('NibbleRaster', "Overwrite")
                
            finalMosaic = "MOSAIC" + "_" + str(counter)
            
            arcpy.MosaicToNewRaster_management(['NibbleRaster', MosaicRaster], self.geoDatabasePath, finalMosaic, "", "8_BIT_UNSIGNED", "", "1", "FIRST", "FIRST")
            self.temporaryFiles.append(finalMosaic)

            # Temporary raster created only in Arc10
            self.temporaryFiles.append('RegionG_Extr' + str(counter))

            MosaicRaster = finalMosaic
            arcpy.CalculateStatistics_management(MosaicRaster)
            arcpy.BuildRasterAttributeTable_management(MosaicRaster, "Overwrite")
            counter += 1
                
        return MosaicRaster

    # Runs Stage 3 of slope script. Converts raster to polygon and adds attribute information
    def Stage3(self, MosaicRaster):

        # Region group, extract by attributes, nibble, and save final raster
        regionGroup = RegionGroup(MosaicRaster, "FOUR", "WITHIN", "NO_LINK")
        regionGroup.save('RegionGroup')
        
        extractRaster = ExtractByAttributes('RegionGroup', self.maxClusterSizeQuery)
        extractRaster.save('ExtractRaster')
        
        nibbleRaster = Nibble(MosaicRaster, 'ExtractRaster')
        nibbleRaster.save("FinalRaster")
        arcpy.BuildRasterAttributeTable_management("FinalRaster", "Overwrite")

        # Convert raster to polygon
        rasterToPoly = os.path.join(arcpy.env.workspace, "rasterToPoly")
        arcpy.RasterToPolygon_conversion("FinalRaster", rasterToPoly, "SIMPLIFY")
        arcpy.AddMessage("Converted Raster to Polygon")
        self.temporaryFiles.append("rasterToPoly")

        # Convert polygon to polyline
        featToLine = os.path.join(arcpy.env.workspace, "polyToLine")
        arcpy.FeatureToLine_management(rasterToPoly, featToLine, '', "ATTRIBUTES")
        self.temporaryFiles.append("polyToLine")

        # Smooth polyline
        smoothPolyline = os.path.join(arcpy.env.workspace, "smoothPolyline")
        arcpy.SmoothLine_cartography(featToLine, smoothPolyline, "PAEK", "40 Meters", "FIXED_CLOSED_ENDPOINT", "NO_CHECK")
        self.temporaryFiles.append("smoothPolyline")

        # Simplify smoothed polyline
        simplifyLine = os.path.join(arcpy.env.workspace, "simplifyLine")
        arcpy.SimplifyLine_cartography(smoothPolyline, simplifyLine, "POINT_REMOVE", "0.5 Meters", "FLAG_ERRORS", "KEEP_COLLAPSED_POINTS", "NO_CHECK")
        self.temporaryFiles.append("simplifyLine")
        self.temporaryFiles.append("simplifyLine_Pnt")

        # Convert simplified polyline to polygon
        featureToPoly = "SlopePolygons"

        # Make direct path to feature because attributes may not be added if this script is run twice?
        featureToPolyPath = os.path.join(arcpy.env.workspace, featureToPoly)
        arcpy.FeatureToPolygon_management(simplifyLine, featureToPolyPath, "", "ATTRIBUTES", "")
        arcpy.DefineProjection_management(featureToPolyPath, self.sr)

        # Add unique ID field to polygon feature class, used for zonal stats 
        arcpy.AddField_management(featureToPolyPath, 'UNIQUEID', 'LONG')
        arcpy.CalculateField_management(featureToPolyPath, 'UNIQUEID', "!OBJECTID!", "PYTHON")

        # Run zonal stats using final raster and feature class
        arcpy.AddMessage("Calculating Feature Class Attributes...")
        zoneDict = self.getZonalStats(featureToPolyPath, "FinalRaster")

        # Add slope class fields to final feature class and update attributes
        reclassDict = self.splitReclassValues(self.reclassValues)
        arcpy.AddField_management(featureToPolyPath, 'SlopeClass', 'LONG')
        arcpy.AddField_management(featureToPolyPath, 'SlopeRange', 'TEXT')
        rows = arcpy.UpdateCursor(featureToPolyPath)
        for row in rows:
            try:
                uniqueID = row.UNIQUEID
                slopeClass = str(zoneDict[uniqueID])
                slopeValues = reclassDict[slopeClass]
                formatSlopeValues = '{0} - {1}'.format(slopeValues[0], slopeValues[1])
                row.setValue('SlopeClass', slopeClass)
                row.setValue('SlopeRange', formatSlopeValues)
                rows.updateRow(row)
            except:
                pass
                    
        del rows

        # Define projection for final raster and feature class
        arcpy.DefineProjection_management("FinalRaster", self.sr)
        
        return featureToPolyPath
        

if __name__ == "__main__":

    # Set the input parameters
    slopeRasterPath = arcpy.GetParameter(0)
    reclassValues = arcpy.GetParameterAsText(1)
    minDelineationSize = arcpy.GetParameter(2)
    geoDatabasePath = arcpy.GetParameterAsText(3)
    deleteTempFiles = arcpy.GetParameter(4)

    #slopeRasterPath = r'C:\Users\Administrator\Desktop\progam_rfp\May_2012\test_steve.gdb\slope2'
    #slopeRasterPath = r'C:\Users\rstanley\Documents\ArcGIS\Default.gdb\slope2stateplane'
    #reclassValues = "0 2 1;2 4 2;4 8 3;8 12 4;12 30 5;30 50 6;50 75 7;75 200 8"
    #minDelineationSize = 5.7
    #geoDatabasePath = r'C:\sxrap.gdb'
    #deleteTempFiles = 1

    # Create an instance of SlopeClass with input parameters
    slopeScript = SlopeClass(slopeRasterPath, reclassValues, minDelineationSize, geoDatabasePath, deleteTempFiles)

    # Call the run method on SlopeClass to run all stages of script
    slopeScript.run()





